from enum import Enum

__all__ = ["EndpointNames"]


class EndpointNames(Enum):
    """Standardised enum of names used in JSON schema filenames for each endpoint."""

    accident_stats = "AccidentStats"
    air_quality = "AirQuality"
    bike_point = "BikePoint"
    crowding = "crowding"
    disruptions_lifts_v2 = "Disruptions-Lifts-v2"
    journey = "Journey"
    line = "Line"
    mode = "Mode"
    occupancy = "occupancy"
    place = "Place"
    road = "Road"
    search = "Search"
    stop_point = "StopPoint"
    vehicle = "Vehicle"



"""Tests for the line API."""

from datetime import datetime, timedelta

from line_data import all_endpoints, snapshot_album
from pytest import mark

from tubeulator import fetch

tested_eps = {
    "all_routes",
    "arrivals_by_ids",
    "arrivals_by_ids_stop",
    "disruption_by_ids",
    "disruption_by_modes",
    "lines_by_ids",
    "lines_by_modes",
    "meta_disruption_categories",
    "meta_modes",
    "meta_service_types",
    "meta_severity",
    "route_by_ids",
    "route_by_modes",
    "route_sequence_by_id_direction",
    "search_lines_routes",
    "status_by_ids",
    "status_by_ids_dates",
    "status_by_modes",
    "status_by_severity",
    "stop_points_by_id",
    "timetable_by_id_from_stop",
    "timetable_by_id_from_to_stop",
}
untested_eps = {
    "forward_requests",
}


def test_line_endpoints():
    assert list(vars(fetch.line)) == all_endpoints
    assert set(vars(fetch.line)) == tested_eps.union(untested_eps)


def test_arrivals_by_ids():
    arrivals_by_ids = fetch.line.arrivals_by_ids(ids="waterloo-city")
    for arrival in arrivals_by_ids:
        assert arrival


def test_all_routes():
    all_routes = fetch.line.all_routes()
    assert all_routes


def test_arrivals_by_ids_stop():
    arrivals_by_ids_stop = fetch.line.arrivals_by_ids_stop(
        ids="waterloo-city",
        stopPointId="940GZZLUASL",
    )
    for arrival in arrivals_by_ids_stop:
        assert arrival


def test_disruption_by_ids():
    disruption_by_ids = fetch.line.disruption_by_ids(ids="waterloo-city")
    for disruption in disruption_by_ids:
        assert disruption


def test_disruption_by_modes():
    disruption_by_modes = fetch.line.disruption_by_modes(modes="tube")
    assert disruption_by_modes


@mark.skip()
def test_forward_requests():
    forward_requests = fetch.line.forward_requests()
    assert forward_requests


def test_meta_disruption_categories():
    meta_disruption_categories = fetch.line.meta_disruption_categories()
    assert meta_disruption_categories


def test_meta_service_types():
    meta_service_types = fetch.line.meta_service_types()
    assert meta_service_types


def test_meta_severity():
    meta_severity = fetch.line.meta_severity()
    assert meta_severity


def test_route_by_ids():
    route_by_ids = fetch.line.route_by_ids(ids="waterloo-city")
    assert route_by_ids


def test_route_by_modes():
    route_by_modes = fetch.line.route_by_modes(modes="tube")
    assert route_by_modes


def test_route_sequence_by_id_direction():
    route_sequence_by_id_direction = fetch.line.route_sequence_by_id_direction(
        id="waterloo-city",
        direction="outbound",
    )
    assert route_sequence_by_id_direction


def test_search_lines_routes():
    search_lines_routes = fetch.line.search_lines_routes(query="Downing Street")
    assert search_lines_routes


def test_status_by_ids():
    status_by_ids = fetch.line.status_by_ids(ids="waterloo-city")
    assert status_by_ids


def test_status_by_ids_dates():
    now = datetime.now()
    start = (now + timedelta(days=6)).strftime("%Y-%m-%dT%H:%M:%S-00:00")
    end = (now + timedelta(days=7)).strftime("%Y-%m-%dT%H:%M:%S-00:00")
    status_by_ids_dates = fetch.line.status_by_ids_dates(
        ids="waterloo-city",
        startDate=start,
        endDate=end,
    )
    assert status_by_ids_dates


def test_status_by_modes():
    status_by_modes = fetch.line.status_by_modes(modes="tube")
    assert status_by_modes


def test_status_by_severity():
    status_by_severity = fetch.line.status_by_severity(severity=0)
    assert status_by_severity


def test_stop_points_by_id():
    stop_points_by_id = fetch.line.stop_points_by_id(id="waterloo-city")
    assert stop_points_by_id


def test_timetable_by_id_from_stop():
    timetable_by_id_from_stop = fetch.line.timetable_by_id_from_stop(
        id="piccadilly",
        fromStopPointId="940GZZLUASL",
    )
    assert timetable_by_id_from_stop


def test_timetable_by_id_from_to_stop():
    """Arsenal to Piccadilly Circus, via Piccadilly line."""
    timetable_by_id_from_to_stop = fetch.line.timetable_by_id_from_to_stop(
        id="piccadilly",
        fromStopPointId="940GZZLUASL",
        toStopPointId="940GZZLUPCC",
    )
    assert timetable_by_id_from_to_stop


@mark.parametrize("expected", [snapshot_album["meta_modes"]])
def test_meta_modes(expected):
    modes = fetch.line.meta_modes()
    assert [m.model_dump(mode="json") for m in modes] == expected


@mark.parametrize("line_id", ["waterloo-city"])
def test_lines_by_ids(line_id):
    """Relies on correct codegen of arrays (previously had a bug)."""
    expected = snapshot_album["lines_by_ids"][line_id]
    lines = fetch.line.lines_by_ids(line_id)
    assert [l.model_dump(mode="json") for l in lines] == expected


@mark.parametrize("mode_id,expected", [("tube", snapshot_album["lines_by_modes"])])
def test_lines_by_modes(mode_id, expected):
    lines = fetch.line.lines_by_modes(mode_id)
    assert [l.model_dump(mode="json") for l in lines] == expected



from __future__ import annotations
import json
from datetime import datetime
from dataclasses import dataclass, field
from tubeulator.utils.string_conv import to_camel_case
from pydantic import AliasGenerator, BaseModel, ConfigDict, PrivateAttr
from ..utils.paths import DtoEnum

class Bay(BaseModel):
    """
    Autogenerated from occupancy::Tfl.Api.Presentation.Entities.Bay
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    BayType: str = None
    BayCount: int = None
    Free: int = None
    Occupied: int = None
    _source_schema_name: str = PrivateAttr(default='occupancy')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.Bay')


BayModel = Bay


class CarParkOccupancy(BaseModel):
    """
    Autogenerated from occupancy::Tfl.Api.Presentation.Entities.CarParkOccupancy
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    Bays: list["BayModel"] = []
    Name: str = None
    CarParkDetailsUrl: str = None
    _source_schema_name: str = PrivateAttr(default='occupancy')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.CarParkOccupancy')


CarParkOccupancyModel = CarParkOccupancy


class ChargeConnectorOccupancy(BaseModel):
    """
    Autogenerated from occupancy::Tfl.Api.Presentation.Entities.ChargeConnectorOccupancy
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: int = None
    SourceSystemPlaceId: str = None
    Status: str = None
    _source_schema_name: str = PrivateAttr(default='occupancy')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.ChargeConnectorOccupancy')


ChargeConnectorOccupancyModel = ChargeConnectorOccupancy


class BikePointOccupancy(BaseModel):
    """
    Autogenerated from occupancy::Tfl.Api.Presentation.Entities.BikePointOccupancy
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    Name: str = None
    BikesCount: int = None
    EmptyDocks: int = None
    TotalDocks: int = None
    _source_schema_name: str = PrivateAttr(default='occupancy')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.BikePointOccupancy')


BikePointOccupancyModel = BikePointOccupancy


class Deserialisers(DtoEnum):
    Tfl__Api__Presentation__Entities__Bay = Bay
    Tfl__Api__Presentation__Entities__CarParkOccupancy = CarParkOccupancy
    Tfl__Api__Presentation__Entities__ChargeConnectorOccupancy = ChargeConnectorOccupancy
    Tfl__Api__Presentation__Entities__BikePointOccupancy = BikePointOccupancy


import random

from .combine import load_station_points_with_lines

__all__ = ["show_station_points_map"]


def get_line_colour(line):
    line2colour = {
        "london-cable-car": "#D799AF",  # Light purple
        "dlr": "#00AFAD",  # Turquoise
        "thameslink": "#0019A8",  # Dark blue
        "national-rail": "#A0A5A9",  # Silver
        "london-overground": "#E86A10",  # Orange
        "elizabeth": "#5A4799",  # Purple
        "tram": "#00BD19",  # Green
        "bakerloo": "#B36305",  # Brown
        "central": "#E32017",  # Red
        "circle": "#FFD300",  # Yellow
        "district": "#00782A",  # Green
        "hammersmith-city": "#F3A9BB",  # Pink
        "jubilee": "#868F98",  # Grey
        "metropolitan": "#9B0056",  # Magenta
        "northern": "#000000",  # Black
        "piccadilly": "#003688",  # Dark blue
        "victoria": "#0098D4",  # Light blue
        "waterloo-city": "#93CEBA",  # Pale turquoise
    }
    return line2colour[line]


def show_station_points_map(zoom: int = 12) -> None:
    """Create a map centered on London and plot all station points."""
    # TODO: ideally handle this using a Pydantic model `ImportPath`
    try:
        import folium as fl
    except ImportError:
        raise ImportError(
            "Please install the folium package: `pip install tubeulator[folium]`.",
        )
    centre_of_london = [51.5074, -0.1278]
    map_style = "cartodbpositron"
    london_map = fl.Map(location=centre_of_london, zoom_start=zoom, tiles=map_style)
    stn_points = load_station_points_with_lines()
    for stp in stn_points.to_dicts():
        coord = [stp["Lat"], stp["Lon"]]
        pt_name = stp["FriendlyName"]
        stn_name = stp["StationName"]
        lines = ", ".join([l.title() for l in stp["Lines"]])
        popup_text = f"{pt_name} at {stn_name} ({lines})"
        marker_colour = get_line_colour(random.choice(stp["Lines"]))
        styled_icon = fl.Icon(color="white", icon_color=marker_colour)
        fl.Marker(location=coord, popup=popup_text, icon=styled_icon).add_to(london_map)
    london_map.show_in_browser()
    return



"""Accessed via dynamic method resolution under `tubeulator.fetch.vehicle`.

!!! example "Example: `tubeulator.fetch.vehicle.vehicle_arrivals()`"

    ```py
    >>> arrivals = fetch.vehicle.vehicle_arrivals(ids="100")
    >>> len(arrivals)
    9
    >>> print(arrivals[0].model_dump_json(indent=2))
    {
      "Id": "1163507335",
      "OperationType": 1,
      "VehicleId": "100",
      "NaptanId": "940GZZLUHGT",
      "StationName": "Highgate Underground Station",
      "LineId": "northern",
      "LineName": "Northern",
      "PlatformName": "Northbound - Platform 1",
      "Direction": "outbound",
      "Bearing": "",
      "DestinationNaptanId": "940GZZLUHBT",
      "DestinationName": "High Barnet Underground Station",
      "Timestamp": "2024-05-21T23:20:28.040799Z",
      "TimeToStation": 32,
      "CurrentLocation": "Between Archway and Highgate",
      "Towards": "High Barnet via Bank",
      "ExpectedArrival": "2024-05-21T23:21:00Z",
      "TimeToLive": "2024-05-21T23:21:00Z",
      "ModeName": "tube",
      "Timing": {
        "CountdownServerAdjustment": "00:00:00",
        "Source": "0001-01-01T00:00:00",
        "Insert": "0001-01-01T00:00:00",
        "Read": "2024-05-21T23:20:03.635000Z",
        "Sent": "2024-05-21T23:20:28Z",
        "Received": "0001-01-01T00:00:00"
      }
    }
    ```
"""

from .types import RouteEnum


class VehicleEndpointRoutes(RouteEnum):
    emission_surcharge = "/EmissionSurcharge"
    """Gets the Emissions Surcharge compliance for the Vehicle"""
    ulez_compliance = "/UlezCompliance"
    """Gets the Ulez Surcharge compliance for the Vehicle"""
    vehicle_arrivals = "/{ids}/Arrivals"
    """Gets the predictions for a given list of vehicle Id's."""



from inline_snapshot import snapshot

import tubeulator


def test_interface():
    """This list was used to generate the `tests/apis/` test files.

    If the snapshot changes (a new interface is generated) then add test coverage.
    """
    interfaces = snapshot(
        [
            "accident_stats",
            "air_quality",
            "bike_point",
            "crowding",
            "disruptions_lifts_v2",
            "journey",
            "line",
            "mode",
            "occupancy",
            "place",
            "road",
            "search",
            "stop_point",
            "vehicle",
        ],
    )
    list(vars(tubeulator.fetch)) == interfaces



from inline_snapshot import snapshot

all_endpoints = snapshot(
    [
        "bike_points",
        "car_park",
        "car_park_id",
        "charge_connector",
        "charge_connector_ids",
    ],
)



from inline_snapshot import snapshot

all_endpoints = snapshot(
    [
        "forward_requests",
        "meta_disruption_categories",
        "meta_modes",
        "meta_service_types",
        "meta_severity",
        "lines_by_modes",
        "disruption_by_modes",
        "route_by_modes",
        "status_by_modes",
        "all_routes",
        "search_lines_routes",
        "status_by_severity",
        "lines_by_ids",
        "arrivals_by_ids",
        "arrivals_by_ids_stop",
        "disruption_by_ids",
        "route_by_ids",
        "status_by_ids",
        "status_by_ids_dates",
        "route_sequence_by_id_direction",
        "stop_points_by_id",
        "timetable_by_id_from_stop",
        "timetable_by_id_from_to_stop",
    ],
)

meta_modes_snapshot = snapshot(
    [
        {
            "IsTflService": True,
            "IsFarePaying": True,
            "IsScheduledService": True,
            "ModeName": "bus",
        },
        {
            "IsTflService": True,
            "IsFarePaying": True,
            "IsScheduledService": True,
            "ModeName": "cable-car",
        },
        {
            "IsTflService": False,
            "IsFarePaying": True,
            "IsScheduledService": True,
            "ModeName": "coach",
        },
        {
            "IsTflService": False,
            "IsFarePaying": False,
            "IsScheduledService": False,
            "ModeName": "cycle",
        },
        {
            "IsTflService": True,
            "IsFarePaying": True,
            "IsScheduledService": False,
            "ModeName": "cycle-hire",
        },
        {
            "IsTflService": True,
            "IsFarePaying": True,
            "IsScheduledService": True,
            "ModeName": "dlr",
        },
        {
            "IsTflService": True,
            "IsFarePaying": True,
            "IsScheduledService": True,
            "ModeName": "elizabeth-line",
        },
        {
            "IsTflService": False,
            "IsFarePaying": False,
            "IsScheduledService": False,
            "ModeName": "interchange-keep-sitting",
        },
        {
            "IsTflService": False,
            "IsFarePaying": False,
            "IsScheduledService": False,
            "ModeName": "interchange-secure",
        },
        {
            "IsTflService": False,
            "IsFarePaying": True,
            "IsScheduledService": True,
            "ModeName": "national-rail",
        },
        {
            "IsTflService": True,
            "IsFarePaying": True,
            "IsScheduledService": True,
            "ModeName": "overground",
        },
        {
            "IsTflService": True,
            "IsFarePaying": True,
            "IsScheduledService": True,
            "ModeName": "replacement-bus",
        },
        {
            "IsTflService": True,
            "IsFarePaying": True,
            "IsScheduledService": True,
            "ModeName": "river-bus",
        },
        {
            "IsTflService": True,
            "IsFarePaying": True,
            "IsScheduledService": True,
            "ModeName": "river-tour",
        },
        {
            "IsTflService": False,
            "IsFarePaying": False,
            "IsScheduledService": False,
            "ModeName": "taxi",
        },
        {
            "IsTflService": True,
            "IsFarePaying": True,
            "IsScheduledService": True,
            "ModeName": "tram",
        },
        {
            "IsTflService": True,
            "IsFarePaying": True,
            "IsScheduledService": True,
            "ModeName": "tube",
        },
        {
            "IsTflService": False,
            "IsFarePaying": False,
            "IsScheduledService": False,
            "ModeName": "walking",
        },
    ],
)
wc_line_snapshot = snapshot(
    [
        {
            "Id": "waterloo-city",
            "Name": "Waterloo & City",
            "ModeName": "tube",
            "Disruptions": [],
            "Created": "2024-05-14T14:35:06.617000Z",
            "Modified": "2024-05-14T14:35:06.617000Z",
            "LineStatuses": [],
            "RouteSections": [],
            "ServiceTypes": [
                {
                    "Name": "Regular",
                    "Uri": "/Line/Route?ids=Waterloo & City&serviceTypes=Regular",
                },
            ],
            "Crowding": {"PassengerFlows": [], "TrainLoadings": []},
        },
    ],
)
tube_line_snapshot = snapshot(
    [
        {
            "Id": "bakerloo",
            "Name": "Bakerloo",
            "ModeName": "tube",
            "Disruptions": [],
            "Created": "2024-05-14T14:35:06.647000Z",
            "Modified": "2024-05-14T14:35:06.647000Z",
            "LineStatuses": [],
            "RouteSections": [],
            "ServiceTypes": [
                {
                    "Name": "Regular",
                    "Uri": "/Line/Route?ids=Bakerloo&serviceTypes=Regular",
                },
            ],
            "Crowding": {"PassengerFlows": [], "TrainLoadings": []},
        },
        {
            "Id": "central",
            "Name": "Central",
            "ModeName": "tube",
            "Disruptions": [],
            "Created": "2024-05-14T14:35:06.630000Z",
            "Modified": "2024-05-14T14:35:06.630000Z",
            "LineStatuses": [],
            "RouteSections": [],
            "ServiceTypes": [
                {
                    "Name": "Regular",
                    "Uri": "/Line/Route?ids=Central&serviceTypes=Regular",
                },
                {"Name": "Night", "Uri": "/Line/Route?ids=Central&serviceTypes=Night"},
            ],
            "Crowding": {"PassengerFlows": [], "TrainLoadings": []},
        },
        {
            "Id": "circle",
            "Name": "Circle",
            "ModeName": "tube",
            "Disruptions": [],
            "Created": "2024-05-14T14:35:06.647000Z",
            "Modified": "2024-05-14T14:35:06.647000Z",
            "LineStatuses": [],
            "RouteSections": [],
            "ServiceTypes": [
                {
                    "Name": "Regular",
                    "Uri": "/Line/Route?ids=Circle&serviceTypes=Regular",
                },
            ],
            "Crowding": {"PassengerFlows": [], "TrainLoadings": []},
        },
        {
            "Id": "district",
            "Name": "District",
            "ModeName": "tube",
            "Disruptions": [],
            "Created": "2024-05-14T14:35:06.630000Z",
            "Modified": "2024-05-14T14:35:06.630000Z",
            "LineStatuses": [],
            "RouteSections": [],
            "ServiceTypes": [
                {
                    "Name": "Regular",
                    "Uri": "/Line/Route?ids=District&serviceTypes=Regular",
                },
            ],
            "Crowding": {"PassengerFlows": [], "TrainLoadings": []},
        },
        {
            "Id": "hammersmith-city",
            "Name": "Hammersmith & City",
            "ModeName": "tube",
            "Disruptions": [],
            "Created": "2024-05-14T14:35:06.647000Z",
            "Modified": "2024-05-14T14:35:06.647000Z",
            "LineStatuses": [],
            "RouteSections": [],
            "ServiceTypes": [
                {
                    "Name": "Regular",
                    "Uri": "/Line/Route?ids=Hammersmith & City&serviceTypes=Regular",
                },
            ],
            "Crowding": {"PassengerFlows": [], "TrainLoadings": []},
        },
        {
            "Id": "jubilee",
            "Name": "Jubilee",
            "ModeName": "tube",
            "Disruptions": [],
            "Created": "2024-05-14T14:35:06.630000Z",
            "Modified": "2024-05-14T14:35:06.630000Z",
            "LineStatuses": [],
            "RouteSections": [],
            "ServiceTypes": [
                {
                    "Name": "Regular",
                    "Uri": "/Line/Route?ids=Jubilee&serviceTypes=Regular",
                },
                {"Name": "Night", "Uri": "/Line/Route?ids=Jubilee&serviceTypes=Night"},
            ],
            "Crowding": {"PassengerFlows": [], "TrainLoadings": []},
        },
        {
            "Id": "metropolitan",
            "Name": "Metropolitan",
            "ModeName": "tube",
            "Disruptions": [],
            "Created": "2024-05-14T14:35:06.647000Z",
            "Modified": "2024-05-14T14:35:06.647000Z",
            "LineStatuses": [],
            "RouteSections": [],
            "ServiceTypes": [
                {
                    "Name": "Regular",
                    "Uri": "/Line/Route?ids=Metropolitan&serviceTypes=Regular",
                },
            ],
            "Crowding": {"PassengerFlows": [], "TrainLoadings": []},
        },
        {
            "Id": "northern",
            "Name": "Northern",
            "ModeName": "tube",
            "Disruptions": [],
            "Created": "2024-05-14T14:35:06.647000Z",
            "Modified": "2024-05-14T14:35:06.647000Z",
            "LineStatuses": [],
            "RouteSections": [],
            "ServiceTypes": [
                {
                    "Name": "Regular",
                    "Uri": "/Line/Route?ids=Northern&serviceTypes=Regular",
                },
                {"Name": "Night", "Uri": "/Line/Route?ids=Northern&serviceTypes=Night"},
            ],
            "Crowding": {"PassengerFlows": [], "TrainLoadings": []},
        },
        {
            "Id": "piccadilly",
            "Name": "Piccadilly",
            "ModeName": "tube",
            "Disruptions": [],
            "Created": "2024-05-14T14:35:06.630000Z",
            "Modified": "2024-05-14T14:35:06.630000Z",
            "LineStatuses": [],
            "RouteSections": [],
            "ServiceTypes": [
                {
                    "Name": "Regular",
                    "Uri": "/Line/Route?ids=Piccadilly&serviceTypes=Regular",
                },
                {
                    "Name": "Night",
                    "Uri": "/Line/Route?ids=Piccadilly&serviceTypes=Night",
                },
            ],
            "Crowding": {"PassengerFlows": [], "TrainLoadings": []},
        },
        {
            "Id": "victoria",
            "Name": "Victoria",
            "ModeName": "tube",
            "Disruptions": [],
            "Created": "2024-05-14T14:35:06.617000Z",
            "Modified": "2024-05-14T14:35:06.617000Z",
            "LineStatuses": [],
            "RouteSections": [],
            "ServiceTypes": [
                {
                    "Name": "Regular",
                    "Uri": "/Line/Route?ids=Victoria&serviceTypes=Regular",
                },
                {"Name": "Night", "Uri": "/Line/Route?ids=Victoria&serviceTypes=Night"},
            ],
            "Crowding": {"PassengerFlows": [], "TrainLoadings": []},
        },
        {
            "Id": "waterloo-city",
            "Name": "Waterloo & City",
            "ModeName": "tube",
            "Disruptions": [],
            "Created": "2024-05-14T14:35:06.617000Z",
            "Modified": "2024-05-14T14:35:06.617000Z",
            "LineStatuses": [],
            "RouteSections": [],
            "ServiceTypes": [
                {
                    "Name": "Regular",
                    "Uri": "/Line/Route?ids=Waterloo & City&serviceTypes=Regular",
                },
            ],
            "Crowding": {"PassengerFlows": [], "TrainLoadings": []},
        },
    ],
)


line_id_album = {
    "waterloo-city": wc_line_snapshot,
}

snapshot_album = {
    "meta_modes": meta_modes_snapshot,
    "lines_by_ids": line_id_album,
    "lines_by_modes": tube_line_snapshot,
}



import polars as pl

from .load import (
    load_platform_services,
    load_platforms,
    load_station_points,
    load_stations,
)

__all__ = [
    "load_platforms_with_stations_and_services",
    "load_lines_by_station",
    "load_stations_by_line",
]


def load_platforms_with_stations_and_services():
    """Annotate platform services with their platform and station info."""
    # There are about 1400 platforms
    platforms = load_platforms()
    # There are around 500 stations
    stations = load_stations()
    # There are around 1600 platform services.
    # There are more than the number of physical platforms because multiple services use
    # the same platform (stopping at different times, e.g. Hammersmith & City/District)
    platform_svcs = load_platform_services()
    # There are the same number of platforms with stations as platforms
    platforms_with_stns = platforms.join(stations, on="StationUniqueId")
    # There are the same number of platforms with stations and services as platform services
    plts_stns_svcs = platforms_with_stns.join(platform_svcs, on="PlatformUniqueId")
    return plts_stns_svcs


def load_lines_by_station():
    pss = load_platforms_with_stations_and_services()
    station2lines = pss.group_by("StationUniqueId").agg(
        [pl.col("Line").unique().alias("Lines")],
    )
    station2lines = station2lines.join(
        pss.select(["StationUniqueId", "StationName"]).unique(),
        on="StationUniqueId",
        how="left",
    )
    return station2lines


def load_stations_by_line():
    station2lines = load_lines_by_station()
    line2stations = (
        station2lines.rename({"Lines": "Line"})
        .explode("Line")
        .group_by("Line")
        .agg(pl.col("StationName"))
    )
    return line2stations


def load_station_points_with_lines():
    stn_points = load_station_points()
    stn2lines = load_lines_by_station()
    stn_points_with_lines = stn_points.join(stn2lines, on="StationUniqueId")
    return stn_points_with_lines



from inline_snapshot import snapshot

all_endpoints = snapshot(
    [
        "all_roads",
        "meta_categories",
        "meta_severities",
        "disruption_by_ids",
        "street_disruption",
        "road_by_ids",
        "road_disruption",
        "road_status",
    ],
)



"""Transport network station topology (i.e. the network without its connectivity)."""

from .combine import (
    load_lines_by_station,
    load_platforms_with_stations_and_services,
    load_stations_by_line,
)
from .load import (
    load_lines,
    load_platform_services,
    load_platforms,
    load_station_points,
    load_stations,
)
from .map import show_station_points_map

__all__ = [
    "load_lines",
    "load_lines_by_station",
    "load_platforms",
    "load_platform_services",
    "load_platforms_with_stations_and_services",
    "load_stations",
    "load_stations_by_line",
    "load_station_points",
    "show_station_points_map",
]



"""Tests for the road API."""

from datetime import datetime, timedelta

from pytest import mark
from road_data import all_endpoints

from tubeulator import fetch

tested_eps = {
    "all_roads",
    "meta_categories",
    "meta_severities",
    "disruption_by_ids",
    "street_disruption",
    "road_by_ids",
    "road_disruption",
    "road_status",
}
untested_eps = set()


def test_road_endpoints():
    assert list(vars(fetch.road)) == all_endpoints
    assert set(vars(fetch.road)) == tested_eps.union(untested_eps)


def test_all_roads():
    all_roads = fetch.road.all_roads()
    assert all_roads


def test_meta_categories():
    meta_categories = fetch.road.meta_categories()
    assert meta_categories


def test_meta_severities():
    meta_severities = fetch.road.meta_severities()
    assert meta_severities


def test_disruption_by_ids():
    """Incident from website, may not be able to keep it?

    20th May 2024: https://tfl.gov.uk/traffic/status/?disruptionIds=TIMS-198697
    """
    disruption_by_ids = fetch.road.disruption_by_ids(disruptionIds="TIMS-198697")
    assert disruption_by_ids


@mark.skip(
    reason="404",
)
def test_street_disruption():
    street_disruption = fetch.road.street_disruption()
    assert street_disruption


@mark.skip(
    reason="Prevented from sending by self.invalidate call (optional parameters)",
)
def test_street_disruption_date_range():
    now = datetime.now()
    start = (now + timedelta(days=-7)).strftime("%Y-%m-%dT%H:%M:%S-00:00")
    end = (now + timedelta(days=7)).strftime("%Y-%m-%dT%H:%M:%S-00:00")
    street_disruption = fetch.road.street_disruption(
        startDate=start,
        endDate=end,
    )
    assert street_disruption


def test_road_by_ids():
    road_by_ids = fetch.road.road_by_ids(ids="A1")
    assert road_by_ids


def test_road_disruption():
    road_disruption = fetch.road.road_disruption(ids="A1")
    for disruption in road_disruption:
        assert disruption


def test_road_status():
    road_status = fetch.road.road_status(ids="A1")
    assert road_status



"""TinyDB implementation for credential storage."""

import os
from functools import lru_cache

from pydantic import BaseModel, Field, ValidationError
from tinydb import TinyDB

from ..utils.logs import set_up_logging
from ..utils.paths import db_path

__all__ = ["check_creds"]

logger = set_up_logging(__name__)


class EnvCredentials(BaseModel):
    app_id: str = Field(validation_alias="TFL_API_KEY_ID")
    primary_key: str = Field(validation_alias="TFL_API_PRIMARY_ACCESS_KEY")
    secondary_key: str = Field(validation_alias="TFL_API_SECONDARY_ACCESS_KEY")


@lru_cache
def check_creds() -> dict[str, str]:
    """Either pick up API credentials from environment variables or retrieve from TinyDB.

    Store the API credentials in a simple TinyDB database if none has been stored before.
    This is used only in interactive sessions when environment variables are not set.
    """
    try:
        # Always try environment variables first
        env_creds = EnvCredentials.model_validate(os.environ)
        credential = env_creds.model_dump()
        logger.debug("Using credentials from environment variables")
        return credential
    except ValidationError:
        # No API keys in environment variables, get it from the TinyDB database
        logger.debug("No API keys in environment variables, checking TinyDB")

        # Ensure database directory exists
        db_path.mkdir(parents=True, exist_ok=True)

        # Open TinyDB file
        db_file = db_path / "credentials.json"
        with TinyDB(db_file) as db:
            credentials = db.all()

            if not credentials:
                # No stored credentials, prompt user to enter them
                logger.info(
                    "No stored credentials found. Please enter your TfL API credentials.",
                )
                default_app_id = "tubeulator"
                app_id = (
                    input(f"Enter your app ID (default: {default_app_id!r}): ")
                    or default_app_id
                )
                primary_key = input("Enter your primary key: ")
                secondary_key = input("Enter your secondary key: ")

                credential = dict(
                    app_id=app_id,
                    primary_key=primary_key,
                    secondary_key=secondary_key,
                )

                # Store the credentials
                db.insert(credential)
                logger.debug("Credentials stored in TinyDB")
            elif len(credentials) > 1:
                # There should only be one credential entry
                logger.warning(
                    f"Multiple credentials found in database: {len(credentials)}",
                )
                # Use the first one
                credential = credentials[0]
            else:
                # Use the stored credentials
                credential = credentials[0]

    return credential



"""Download and store (overwrite) the OpenAPI JSON schemas for TfL APIs."""

import json

import httpx

from ..utils.paths import openapi_schemas_path

__all__ = ["fetch_schema"]


def fetch_schema(schema_name: str, api_version: str) -> str:
    """Fetch a TfL API OpenAPI JSON schema, pretty printing with indent=4."""
    export_url = f"https://api-portal.tfl.gov.uk/developer/apis/{schema_name}"
    headers = {"Accept": "application/vnd.oai.openapi+json"}
    params = {"export": True, "api-version": api_version}
    download = httpx.get(export_url, headers=headers, params=params)
    download.raise_for_status()
    pp_json = json.dumps(download.json(), indent=4)
    return pp_json


def refresh_schema(schema_name: str, api_version: str) -> None:
    """Store a TfL API OpenAPI JSON schema, overwriting if one exists."""
    schema_dir = openapi_schemas_path / schema_name
    schema_file = (schema_dir / schema_name).with_suffix(".json")
    schema_json = fetch_schema(schema_name=schema_name, api_version=api_version)
    if schema_file.exists():
        old_json = schema_file.read_text()
        old_file_size = len(old_json.splitlines())
        new_file_size = len(schema_json.splitlines())
        diff = f"{old_file_size} lines -> {new_file_size} lines"
        is_match = ""
        if old_file_size == new_file_size:
            if old_json.rstrip() == schema_json.rstrip():
                is_match = " -- no change"
        message = f"Writing {schema_file.name} ({diff}){is_match}"
    else:
        message = f"Writing {schema_file.name} (new file)"
    print(message)
    schema_file.write_text(schema_json)
    return



"""Tests for the journey API."""

from journey_data import all_endpoints

from tubeulator import fetch

tested_eps = {"journey_results", "modes"}
untested_eps = {"forward_requests"}


def test_journey_endpoints():
    assert list(vars(fetch.journey)) == all_endpoints
    assert set(vars(fetch.journey)) == tested_eps.union(untested_eps)


def test_journey_results():
    journey_results = fetch.journey.journey_results(
        **{"from": "Tooting Bec Underground", "to": "Tooting Broadway Underground"},
    )
    assert journey_results


def test_modes():
    modes = fetch.journey.modes()
    assert modes



"""Accessed via dynamic method resolution under `tubeulator.fetch.stop_point`.

!!! example "Example: `tubeulator.fetch.stop_point.direction_to()`"

    ```py
    >>> direction_to = fetch.stop_point.direction_to(
    ...         id="940GZZLUASL",
    ...         toStopPointId="940GZZLUHWY",
    ... )
    >>> direction_to
    'inbound'
    ```
"""

from .types import RouteEnum


class StopPointEndpointRoutes(RouteEnum):
    within_radius = "/"
    """Gets a list of StopPoints within {radius} by the specified criteria"""
    forward_requests = "/*"
    """Forwards any remaining requests to the back-end"""
    meta_categories = "/Meta/Categories"
    """Gets the list of available StopPoint additional information categories"""
    meta_modes = "/Meta/Modes"
    """Gets the list of available StopPoint modes"""
    meta_stop_types = "/Meta/StopTypes"
    """Gets the list of available StopPoint types"""
    mode = "/Mode/{modes}"
    """Gets a list of StopPoints filtered by the modes available at that StopPoint."""
    mode_disruption = "/Mode/{modes}/Disruption"
    """Gets a distinct list of disrupted stop points for the given modes"""
    search = "/Search"
    """Search StopPoints by their common name, or their 5-digit Countdown Bus Stop Code."""
    search_query = "/Search/{query}"
    """Search StopPoints by their common name, or their 5-digit Countdown Bus Stop Code."""
    service_types = "/ServiceTypes"
    """Gets the service types for a given stoppoint"""
    sms_id = "/Sms/{id}"
    """Gets a StopPoint for a given sms code."""
    type = "/Type/{types}"
    """Gets all stop points of a given type"""
    type_page = "/Type/{types}/page/{page}"
    """Gets all the stop points of given type(s) with a page number"""
    stop_point_ids = "/{ids}"
    """Gets a list of StopPoints corresponding to the given list of stop ids."""
    stop_point_disruption = "/{ids}/Disruption"
    """Gets all disruptions for the specified StopPointId, plus disruptions for any
    child Naptan records it may have."""
    arrival_departures = "/{id}/ArrivalDepartures"
    """Gets the list of arrival and departure predictions for the given stop point id
    (overground and tfl rail only)"""
    arrivals = "/{id}/Arrivals"
    """Gets the list of arrival predictions for the given stop point id"""
    can_reach_on_line = "/{id}/CanReachOnLine/{lineId}"
    """Gets Stopoints that are reachable from a station/line combination."""
    crowding = "/{id}/Crowding/{line}"
    """Gets all the Crowding data (static) for the StopPointId, plus crowding data
    for a given line and optionally a particular direction."""
    direction_to = "/{id}/DirectionTo/{toStopPointId}"
    """Returns the canonical direction, "inbound" or "outbound", for a given pair of
    stop point Ids in the direction from -&gt; to."""
    route = "/{id}/Route"
    """Returns the route sections for all the lines that service the given stop point
    ids"""
    place_types = "/{id}/placeTypes"
    """Get a list of places corresponding to a given id and place types."""
    car_parks = "/{stopPointId}/CarParks"
    """Get car parks corresponding to the given stop point id."""
    taxi_ranks = "/{stopPointId}/TaxiRanks"
    """Gets a list of taxi ranks corresponding to the given stop point id."""



from inline_snapshot import snapshot

all_endpoints = snapshot(
    [
        "search",
        "bus_schedules",
        "meta_categories",
        "meta_search_providers",
        "meta_sorts",
    ],
)



"""Accessed via dynamic method resolution under `tubeulator.fetch.place`.

!!! example "Example: `tubeulator.fetch.place.meta_categories()`"

    Unfortunately most of these methods don't work. This one isn't much use on its own:

    ```py
    >>> categories = fetch.place.meta_categories()
    >>> print(categories[11].model_dump_json(indent=2))
    {
      "Category": "Load",
      "AvailableKeys": [
        "Pallets/bagged",
        "Liquids/gas",
        "Containers",
        "Bulks",
        "General Cargo",
        "Can accommodate a crane"
      ]
    }
    ```
"""

from .types import RouteEnum


class PlaceEndpointRoutes(RouteEnum):
    places_by_geo_region = "/"
    """Gets the places that lie within a geographic region. The geographic region of interest can either
    be specified by using a lat/lon geo-point and a radius in metres to return places within the locus
    defined by the lat/lon of its centre or alternatively, by the use of a bounding box defined by the
    lat/lon of its north-west and south-east corners. Optionally filters on type and can strip
    properties for a smaller payload."""
    forward_requests = "/*"
    """Forwards any remaining requests to the back-end"""
    meta_categories = "/Meta/Categories"
    """Gets a list of all of the available place property categories and keys."""
    meta_place_types = "/Meta/PlaceTypes"
    """Gets a list of the available types of Place."""
    search_places = "/Search"
    """Gets all places that matches the given query"""
    places_by_type = "/Type/{types}"
    """Gets all places of a given type"""
    place_by_id = "/{id}"
    """Gets the place with the given id."""
    place_by_type_at_coordinates = "/{type}/At/{lat}/{lon}"
    """Gets any places of the given type whose geography intersects the given latitude and longitude. In
    practice this means the Place must be polygonal e.g. a BoroughBoundary."""



"""Tests for the mode API."""

from mode_data import all_endpoints

from tubeulator import fetch

tested_eps = {"active_service_types", "mode_arrivals"}
untested_eps = set()


def test_mode_endpoints():
    assert list(vars(fetch.mode)) == all_endpoints
    assert set(vars(fetch.mode)) == tested_eps.union(untested_eps)


def test_active_service_types():
    active_service_types = fetch.mode.active_service_types()
    assert active_service_types


def test_mode_arrivals():
    mode_arrivals = fetch.mode.mode_arrivals(mode="tube")
    assert mode_arrivals



"""Accessed via dynamic method resolution under `tubeulator.fetch.journey`.

!!! example "Example: `tubeulator.fetch.journey.modes()`"

    ```py
    >>> modes = fetch.journey.modes()
    >>> modes[0]
    Mode(IsTflService=False, IsFarePaying=True, IsScheduledService=False, ModeName='black-cab-as-customer')
    ```
"""

from .types import RouteEnum


class JourneyEndpointRoutes(RouteEnum):
    forward_requests = "/*"
    """Forwards any remaining requests to the back-end"""
    journey_results = "/JourneyResults/{from}/to/{to}"
    """Perform a Journey Planner search from the parameters specified in simple types"""
    modes = "/Meta/Modes"
    """Gets a list of all of the available journey planner modes"""



"""Accessed via dynamic method resolution under `tubeulator.fetch.crowding`.

!!! example "Example: `tubeulator.fetch.crowding.dow_crowding()`"

    This method isn't working right now, but it's called like this:

    ```py
    >>> dow_crowding = fetch.crowding.dow_crowding(Naptan="HUBZWL", DayOfWeek="Mon")
    ```
"""

from .types import RouteEnum


class CrowdingEndpointRoutes(RouteEnum):
    naptan_crowding = "/{Naptan}"
    """Returns crowding information for Naptan"""
    live_crowding = "/{Naptan}/Live"
    """Returns live crowding information for Naptan"""
    dow_crowding = "/{Naptan}/{DayOfWeek}"
    """Returns crowding information for Naptan for Day of Week"""



"""Accessed via dynamic method resolution under `tubeulator.fetch.line`.

!!! example "Example: `tubeulator.fetch.line.meta_modes()`"

    ```py
    >>> modes = fetch.line.meta_modes()
    >>> modes[0]
    Mode(IsTflService=True, IsFarePaying=True, IsScheduledService=True, ModeName='bus')
    ```
"""

from .types import RouteEnum


class LineEndpointRoutes(RouteEnum):
    forward_requests = "/*"
    """Forwards any remaining requests to the back-end"""
    meta_disruption_categories = "/Meta/DisruptionCategories"
    """Gets a list of valid disruption categories"""
    meta_modes = "/Meta/Modes"
    """Gets a list of valid modes"""
    meta_service_types = "/Meta/ServiceTypes"
    """Gets a list of valid ServiceTypes to filter on"""
    meta_severity = "/Meta/Severity"
    """Gets a list of valid severity codes"""
    lines_by_modes = "/Mode/{modes}"
    """Gets lines that serve the given modes."""
    disruption_by_modes = "/Mode/{modes}/Disruption"
    """Get disruptions for all lines of the given modes."""
    route_by_modes = "/Mode/{modes}/Route"
    """Gets all lines and their valid routes for given modes, including the name and id of
    the originating and terminating stops for each route"""
    status_by_modes = "/Mode/{modes}/Status"
    """Gets the line status of for all lines for the given modes"""
    all_routes = "/Route"
    """Get all valid routes for all lines, including the name and id of the originating and
    terminating stops for each route."""
    search_lines_routes = "/Search/{query}"
    """Search for lines or routes matching the query string"""
    status_by_severity = "/Status/{severity}"
    """Gets the line status for all lines with a given severity. A list of valid severity
    codes can be obtained from a call to Line/Meta/Severity"""
    lines_by_ids = "/{ids}"
    """Gets lines that match the specified line ids."""
    arrivals_by_ids = "/{ids}/Arrivals"
    """Get the list of arrival predictions for given line ids based at the given stop"""
    arrivals_by_ids_stop = "/{ids}/Arrivals/{stopPointId}"
    """Get the list of arrival predictions for given line ids based at the given stop"""
    disruption_by_ids = "/{ids}/Disruption"
    """Get disruptions for the given line ids"""
    route_by_ids = "/{ids}/Route"
    """Get all valid routes for given line ids, including the name and id of the
    originating and terminating stops for each route."""
    status_by_ids = "/{ids}/Status"
    """Gets the line status of for given line ids e.g Minor Delays"""
    status_by_ids_dates = "/{ids}/Status/{startDate}/to/{endDate}"
    """Gets the line status for given line ids during the provided dates e.g Minor Delays"""
    route_sequence_by_id_direction = "/{id}/Route/Sequence/{direction}"
    """Gets all valid routes for given line id, including the sequence of stops on each
    route."""
    stop_points_by_id = "/{id}/StopPoints"
    """Gets a list of the stations that serve the given line id"""
    timetable_by_id_from_stop = "/{id}/Timetable/{fromStopPointId}"
    """Gets the timetable for a specified station on the give line"""
    timetable_by_id_from_to_stop = (
        "/{id}/Timetable/{fromStopPointId}/to/{toStopPointId}"
    )
    """Gets the timetable for a specified station on the give line with specified
    destination"""



"""Tests for the stop_point API."""

from inline_snapshot import snapshot
from pytest import mark, raises
from stop_point_data import all_endpoints

from tubeulator import fetch
from tubeulator.exc import RequestError

tested_eps = {
    "search_query",
    "sms_id",
}
untested_eps = {
    "mode",
    "mode_disruption",
    "type",
    "type_page",
    "stop_point_ids",
    "stop_point_disruption",
    "arrival_departures",
    "arrivals",
    "can_reach_on_line",
    "crowding",
    "direction_to",
    "route",
    "place_types",
    "forward_requests",
    "search",
    "service_types",
    "within_radius",
    "meta_categories",
    "meta_modes",
    "meta_stop_types",
    "car_parks",
    "taxi_ranks",
}


def test_stop_point_endpoints():
    assert list(vars(fetch.stop_point)) == all_endpoints
    assert set(vars(fetch.stop_point)) == tested_eps.union(untested_eps)


@mark.skip(reason="404 (endpoint not found)?")
def test_within_radius():
    within_radius = fetch.stop_point.within_radius()
    assert within_radius


@mark.skip(reason="404 (endpoint not found)?")
def test_forward_requests():
    forward_requests = fetch.stop_point.forward_requests()
    assert forward_requests


def test_meta_categories():
    meta_categories = fetch.stop_point.meta_categories()
    assert meta_categories


def test_meta_modes():
    meta_modes = fetch.stop_point.meta_modes()
    assert meta_modes


def test_meta_stop_types():
    meta_stop_types = fetch.stop_point.meta_stop_types()
    assert meta_stop_types


@mark.skip(
    reason="Times out (big and slow?), 'bus' mode is rejected without pagination",
)
def test_mode():
    mode = fetch.stop_point.mode(modes="tube")
    assert mode


def test_mode_disruption():
    mode_disruption = fetch.stop_point.mode_disruption(modes="overground")
    assert mode_disruption


@mark.skip(reason="404 (endpoint not found)?")
def test_search():
    search = fetch.stop_point.search()
    assert search


@mark.parametrize(
    "query_input,result_size",
    [("Bromley-By-Bow", 1), ("Euston", snapshot(6))],
)
def test_search_query(query_input, result_size):
    search_query = fetch.stop_point.search_query(query=query_input)
    assert search_query.Total == result_size


@mark.skip(reason="Returns a HTTP error")
def test_service_types():
    service_types = fetch.stop_point.service_types()
    assert service_types


def test_sms_id():
    """Returns a 302 (redirect)"""
    with raises(RequestError):
        fetch.stop_point.sms_id(id=73241)


def test_type():
    type = fetch.stop_point.type(types="TransportInterchange")
    assert type


def test_type_page():
    type_page = fetch.stop_point.type_page(types="TransportInterchange", page=1)
    assert type_page


def test_stop_point_ids():
    stop_point_ids = fetch.stop_point.stop_point_ids(ids="HUBWAT")
    assert stop_point_ids


def test_stop_point_disruption():
    stop_point_disruption = fetch.stop_point.stop_point_disruption(ids="HUBWAT")
    for disruption in stop_point_disruption:
        assert disruption


@mark.skip(reason="Says lineIds is not in endpoint signature but schema shows it is")
def test_arrival_departures():
    arrival_departures = fetch.stop_point.arrival_departures(
        id="HUBWAT",
        lineIds="tfl-rail",
    )
    assert arrival_departures


def test_arrivals():
    arrivals = fetch.stop_point.arrivals(id="HUBWAT")
    for arrival in arrivals:
        assert arrival


@mark.skip(
    reason="Non-bool status: 'Unknown'. Should be enum not bool? Schema line 2214",
)
def test_can_reach_on_line():
    can_reach_on_line = fetch.stop_point.can_reach_on_line(
        id="940GZZLUASL",
        lineId="Piccadilly",
    )
    assert can_reach_on_line


@mark.skip(reason="404 (endpoint not found)?")
def test_crowding():
    crowding = fetch.stop_point.crowding(id="940GZZLUASL", line="Piccadilly")
    assert crowding


def test_direction_to():
    direction_to = fetch.stop_point.direction_to(
        id="940GZZLUASL",
        toStopPointId="940GZZLUHWY",
    )
    assert direction_to


def test_route():
    route = fetch.stop_point.route(id="940GZZLUASL")
    assert route


@mark.skip(reason="404 (endpoint not found)?")
def test_place_types():
    place_types = fetch.stop_point.place_types(id="940GZZLUASL")
    assert place_types


def test_car_parks():
    car_parks = fetch.stop_point.car_parks(stopPointId="HUBWAT")
    for parking in car_parks:
        assert parking


def test_taxi_ranks():
    taxi_ranks = fetch.stop_point.taxi_ranks(stopPointId="HUBWAT")
    for taxis in taxi_ranks:
        assert taxis



"""Tests for the disruptions_lifts_v2 API."""

from disruptions_lifts_v2_data import all_endpoints
from pytest import mark

from tubeulator import fetch

tested_eps = set()
untested_eps = {"all"}


def test_disruptions_lifts_v2_endpoints():
    assert list(vars(fetch.disruptions_lifts_v2)) == all_endpoints
    assert set(vars(fetch.disruptions_lifts_v2)) == tested_eps.union(
        untested_eps,
    )


@mark.skip(reason="404")
def test_disruptions_lifts_v2():
    disruptions = fetch.disruptions_lifts_v2.all()
    assert disruptions



from __future__ import annotations

from enum import Enum

from patito import Model


class LocationTypeEnum(Enum):
    Stop = "0"
    Station = "1"
    # EntranceOrExit = "2"  # Not used
    GenericNode = "3"
    # BoardingArea = "4"  # Not used


class GTFSStop(Model):
    stop_id: str
    stop_code: str | None = None
    stop_name: str
    stop_desc: str | None = None
    stop_lat: float | None = None
    stop_lon: float | None = None
    location_type: int  # LocationTypeEnum
    parent_station: str | None = None
    level_id: str | None = None
    platform_code: str | None = None



"""Accessed via dynamic method resolution under `tubeulator.fetch.mode`.

!!! example "Example: `tubeulator.fetch.mode.mode_arrivals()`"

    ```py
    >>> arrivals = tubeulator.fetch.mode.mode_arrivals(mode="tube")
    >>> print(arrivals[0].model_dump_json(indent=2))
    {
      "Id": "548560311",
      "OperationType": 1,
      "VehicleId": "305",
      "NaptanId": "940GZZLUSWK",
      "StationName": "Southwark Underground Station",
      "LineId": "jubilee",
      "LineName": "Jubilee",
      "PlatformName": "Westbound - Platform 1",
      "Direction": "inbound",
      "Bearing": "",
      "DestinationNaptanId": "940GZZLUSTM",
      "DestinationName": "Stanmore Underground Station",
      "Timestamp": "2024-05-21T22:59:57.338367Z",
      "TimeToStation": 16,
      "CurrentLocation": "Between London Bridge and Southwark",
      "Towards": "Stanmore",
      "ExpectedArrival": "2024-05-21T23:00:13Z",
      "TimeToLive": "2024-05-21T23:00:13Z",
      "ModeName": "tube",
      "Timing": {
        "CountdownServerAdjustment": "00:00:00",
        "Source": "0001-01-01T00:00:00",
        "Insert": "0001-01-01T00:00:00",
        "Read": "2024-05-21T22:59:28.617000Z",
        "Sent": "2024-05-21T22:59:57Z",
        "Received": "0001-01-01T00:00:00"
      }
    }
    ```
"""

from .types import RouteEnum


class ModeEndpointRoutes(RouteEnum):
    active_service_types = "/ActiveServiceTypes"
    """Returns the service type active for a mode. Currently only supports tube"""
    mode_arrivals = "/{mode}/Arrivals"
    """Gets the next arrival predictions for all stops of a given mode"""



from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path

from .. import __path__

__all__ = [
    "pkg_path",
    "data_path",
    "db_path",
    "openapi_schemas_path",
    "openapi_unified_path",
    "stationdata_path",
    "stationdata_gtfs_path",
    "stationdata_detailed_path",
    "RefPath",
    "SchemaPath",
    "to_enum_friendly_str",
    "from_enum_friendly_str",
    "DtoEnum",
]

pkg_path = Path(*__path__)
data_path = pkg_path / "data"
db_path = data_path / "db"
openapi_schemas_path = data_path / "openapi"
openapi_unified_path = data_path / "openapi_unified"
stationdata_path = data_path / "stationdata"
stationdata_gtfs_path = stationdata_path / "gtfs"
stationdata_detailed_path = stationdata_path / "detailed"


@dataclass
class RefPath:
    path: str = field(repr=False)
    name: str = field(init=False)

    def __post_init__(self):
        self.name = Path(self.path).name


@dataclass
class SchemaPath:
    source: dict[str, str] = field(repr=False)
    ref: RefPath = field(init=False)

    def __post_init__(self):
        ref = self.source.get("items", {}).get("$ref")
        self.ref = RefPath(ref)


def to_enum_friendly_str(name: str) -> str:
    return name.replace("-", "_").replace(".", "__")


def from_enum_friendly_str(name: str) -> str:
    return name.replace("_", "-").replace("__", ".")


class DtoEnum(Enum):
    @classmethod
    def select_component(cls, component_name: str):
        return cls[to_enum_friendly_str(component_name)]



"""Tests for the crowding API."""

from crowding_data import all_endpoints
from pytest import mark

from tubeulator import fetch

tested_eps = {"naptan_crowding", "live_crowding", "dow_crowding"}
untested_eps = set()


def test_crowding_endpoints():
    assert list(vars(fetch.crowding)) == all_endpoints
    assert set(vars(fetch.crowding)) == tested_eps.union(untested_eps)


@mark.skip(reason="KeyError: 'content' (in the response_refpath method)")
def test_naptan_crowding():
    naptan_crowding = fetch.crowding.naptan_crowding(Naptan="HUBZWL")
    assert naptan_crowding


@mark.skip(reason="KeyError: 'content' (in the response_refpath method)")
def test_live_crowding():
    live_crowding = fetch.crowding.live_crowding(Naptan="HUBZWL")
    assert live_crowding


@mark.skip(reason="KeyError: 'content' (in the response_refpath method)")
def test_dow_crowding():
    dow_crowding = fetch.crowding.dow_crowding(Naptan="HUBZWL", DayOfWeek="Mon")
    assert dow_crowding



"""Accessed via dynamic method resolution under `tubeulator.fetch.occupancy`.

!!! example "Example: `tubeulator.fetch.occupancy.meta_modes()`"

    ```py
    >>> ccs = fetch.occupancy.charge_connector_ids(ids="ChargePointCM-BPT62200886-143260")
    >>> len(ccs)
    1
    >>> print(ccs[0].model_dump_json(indent=2))
    {
      "Id": 62824,
      "SourceSystemPlaceId": "ChargePointCM-BPT62200886-143260",
      "Status": "Available"
    }
    ```
"""

from .types import RouteEnum


class OccupancyEndpointRoutes(RouteEnum):
    bike_points = "/BikePoints/{ids}"
    """Get the occupancy for bike points."""
    car_park = "/CarPark"
    """Gets the occupancy for all car parks that have occupancy data"""
    car_park_id = "/CarPark/{id}"
    """Gets the occupancy for a car park with a given id"""
    charge_connector = "/ChargeConnector"
    """Gets the occupancy for all charge connectors"""
    charge_connector_ids = "/ChargeConnector/{ids}"
    """Gets the occupancy for a charge connectors with a given id (sourceSystemPlaceId)"""



"""Accessed via dynamic method resolution under `tubeulator.fetch.disruption_lifts`.

!!! example "Example: `fetch.disruptions_lifts_v2.all()`"

    Somewhat ironically, the lift disruptions API is out of order (eturns a 404 when called).

    ```py
    >>> disruptions = fetch.disruptions_lifts_v2.all()
    ```
"""

from .types import RouteEnum


class DisruptionsLiftsEndpointRoutes(RouteEnum):
    all = "/"
    """List of all currently disrupted lift routes"""



from inline_snapshot import snapshot

all_endpoints = snapshot(
    [
        "places_by_geo_region",
        "forward_requests",
        "meta_categories",
        "meta_place_types",
        "search_places",
        "places_by_type",
        "place_by_id",
        "place_by_type_at_coordinates",
    ],
)



"""Tests for the vehicle API."""

from pytest import mark
from vehicle_data import all_endpoints

from tubeulator import fetch

tested_eps = {"vehicle_arrivals"}
untested_eps = {"emission_surcharge", "ulez_compliance"}


def test_vehicle_endpoints():
    assert list(vars(fetch.vehicle)) == all_endpoints
    assert set(vars(fetch.vehicle)) == tested_eps.union(untested_eps)


@mark.skip(reason="404 (endpoint not found)?")
def test_emission_surcharge():
    emission_surcharge = fetch.vehicle.emission_surcharge()
    assert emission_surcharge


@mark.skip(reason="404 (endpoint not found)?")
def test_ulez_compliance():
    ulez_compliance = fetch.vehicle.ulez_compliance()
    assert ulez_compliance


def test_vehicle_arrivals():
    vehicle_arrivals = fetch.vehicle.vehicle_arrivals(ids="100")
    assert vehicle_arrivals



"""Accessed via dynamic method resolution under `tubeulator.fetch.air_quality_data`.

!!! example "Example: `tubeulator.fetch.air_quality.air_quality_data()`"

    This API doesn't seem to be working...

    ```py
    >>> data = fetch.air_quality.air_quality_data()
    >>> data
    Object()
    ```
"""

from .types import RouteEnum


class AirQualityEndpointRoutes(RouteEnum):
    air_quality_data = "/"
    """Gets air quality data feed"""



from __future__ import annotations
import json
from datetime import datetime
from dataclasses import dataclass, field
from tubeulator.utils.string_conv import to_camel_case
from pydantic import AliasGenerator, BaseModel, ConfigDict, PrivateAttr
from ..utils.paths import DtoEnum

class StopPointCategory(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.StopPointCategory
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Category: str = None
    AvailableKeys: list[str] = []
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.StopPointCategory')


StopPointCategoryModel = StopPointCategory


class Mode(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.Mode
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    IsTflService: bool = None
    IsFarePaying: bool = None
    IsScheduledService: bool = None
    ModeName: str = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.Mode')


ModeModel = Mode


class PassengerFlow(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.PassengerFlow
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    TimeSlice: str = None
    Value: int = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.PassengerFlow')


PassengerFlowModel = PassengerFlow


class TrainLoading(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.TrainLoading
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Line: str = None
    LineDirection: str = None
    PlatformDirection: str = None
    Direction: str = None
    NaptanTo: str = None
    TimeSlice: str = None
    Value: int = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.TrainLoading')


TrainLoadingModel = TrainLoading


class Crowding(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.Crowding
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    PassengerFlows: list["PassengerFlowModel"] = []
    TrainLoadings: list["TrainLoadingModel"] = []
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.Crowding')


CrowdingModel = Crowding


class Identifier(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.Identifier
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    Name: str = None
    Uri: str = None
    FullName: str = None
    Type: str = None
    Crowding: CrowdingModel = None
    RouteType: str = None
    Status: str = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.Identifier')


IdentifierModel = Identifier


class LineGroup(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.LineGroup
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    NaptanIdReference: str = None
    StationAtcoCode: str = None
    LineIdentifier: list[str] = []
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.LineGroup')


LineGroupModel = LineGroup


class LineModeGroup(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.LineModeGroup
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    ModeName: str = None
    LineIdentifier: list[str] = []
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.LineModeGroup')


LineModeGroupModel = LineModeGroup


class AdditionalProperties(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.AdditionalProperties
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Category: str = None
    Key: str = None
    SourceSystemKey: str = None
    Value: str = None
    Modified: datetime = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.AdditionalProperties')


AdditionalPropertiesModel = AdditionalProperties


class Place(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.Place
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    Url: str = None
    CommonName: str = None
    Distance: float = None
    PlaceType: str = None
    AdditionalProperties: list["AdditionalPropertiesModel"] = []
    Children: list["PlaceModel"] = []
    ChildrenUrls: list[str] = []
    Lat: float = None
    Lon: float = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.Place')


PlaceModel = Place


class StopPoint(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.StopPoint
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    NaptanId: str = None
    PlatformName: str = None
    Indicator: str = None
    StopLetter: str = None
    Modes: list[str] = []
    IcsCode: str = None
    SmsCode: str = None
    StopType: str = None
    StationNaptan: str = None
    AccessibilitySummary: str = None
    HubNaptanCode: str = None
    Lines: list["IdentifierModel"] = []
    LineGroup: list["LineGroupModel"] = []
    LineModeGroups: list["LineModeGroupModel"] = []
    FullName: str = None
    NaptanMode: str = None
    Status: bool = None
    Id: str = None
    Url: str = None
    CommonName: str = None
    Distance: float = None
    PlaceType: str = None
    AdditionalProperties: list["AdditionalPropertiesModel"] = []
    Children: list["PlaceModel"] = []
    ChildrenUrls: list[str] = []
    Lat: float = None
    Lon: float = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.StopPoint')


StopPointModel = StopPoint


class LineServiceTypeInfo(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.LineServiceTypeInfo
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Name: str = None
    Uri: str = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.LineServiceTypeInfo')


LineServiceTypeInfoModel = LineServiceTypeInfo


class LineSpecificServiceType(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.LineSpecificServiceType
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    ServiceType: LineServiceTypeInfoModel = None
    StopServesServiceType: bool = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.LineSpecificServiceType')


LineSpecificServiceTypeModel = LineSpecificServiceType


class LineServiceType(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.LineServiceType
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    LineName: str = None
    LineSpecificServiceTypes: list["LineSpecificServiceTypeModel"] = []
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.LineServiceType')


LineServiceTypeModel = LineServiceType


class PredictionTiming(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.PredictionTiming
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    CountdownServerAdjustment: str = None
    Source: datetime = None
    Insert: datetime = None
    Read: datetime = None
    Sent: datetime = None
    Received: datetime = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.PredictionTiming')


PredictionTimingModel = PredictionTiming


class Prediction(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.Prediction
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    OperationType: int = None
    VehicleId: str = None
    NaptanId: str = None
    StationName: str = None
    LineId: str = None
    LineName: str = None
    PlatformName: str = None
    Direction: str = None
    Bearing: str = None
    DestinationNaptanId: str = None
    DestinationName: str = None
    Timestamp: datetime = None
    TimeToStation: int = None
    CurrentLocation: str = None
    Towards: str = None
    ExpectedArrival: datetime = None
    TimeToLive: datetime = None
    ModeName: str = None
    Timing: PredictionTimingModel = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.Prediction')


PredictionModel = Prediction


class ArrivalDeparture(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.ArrivalDeparture
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    PlatformName: str = None
    DestinationNaptanId: str = None
    DestinationName: str = None
    NaptanId: str = None
    StationName: str = None
    EstimatedTimeOfArrival: datetime = None
    ScheduledTimeOfArrival: datetime = None
    EstimatedTimeOfDeparture: datetime = None
    ScheduledTimeOfDeparture: datetime = None
    MinutesAndSecondsToArrival: str = None
    MinutesAndSecondsToDeparture: str = None
    Cause: str = None
    DepartureStatus: str = None
    Timing: PredictionTimingModel = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.ArrivalDeparture')


ArrivalDepartureModel = ArrivalDeparture


class StopPointRouteSection(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.StopPointRouteSection
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    NaptanId: str = None
    LineId: str = None
    Mode: str = None
    ValidFrom: datetime = None
    ValidTo: datetime = None
    Direction: str = None
    RouteSectionName: str = None
    LineString: str = None
    IsActive: bool = None
    ServiceType: str = None
    VehicleDestinationText: str = None
    DestinationName: str = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.StopPointRouteSection')


StopPointRouteSectionModel = StopPointRouteSection


class DisruptedPoint(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.DisruptedPoint
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    AtcoCode: str = None
    FromDate: datetime = None
    ToDate: datetime = None
    Description: str = None
    CommonName: str = None
    Type: str = None
    Mode: str = None
    StationAtcoCode: str = None
    Appearance: str = None
    AdditionalInformation: str = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.DisruptedPoint')


DisruptedPointModel = DisruptedPoint


class StopPointsResponse(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.StopPointsResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    CentrePoint: list[float] = []
    StopPoints: list["StopPointModel"] = []
    PageSize: int = None
    Total: int = None
    Page: int = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.StopPointsResponse')


StopPointsResponseModel = StopPointsResponse


class SearchMatch(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.SearchMatch
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    Url: str = None
    Name: str = None
    Lat: float = None
    Lon: float = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.SearchMatch')


SearchMatchModel = SearchMatch


class SearchResponse(BaseModel):
    """
    Autogenerated from StopPoint::Tfl.Api.Presentation.Entities.SearchResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Query: str = None
    From: int = None
    Page: int = None
    PageSize: int = None
    Provider: str = None
    Total: int = None
    Matches: list["SearchMatchModel"] = []
    MaxScore: float = None
    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.SearchResponse')


SearchResponseModel = SearchResponse


class Object(BaseModel):
    """
    Autogenerated from StopPoint::System.Object
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='StopPoint')
    _component_schema_name: str = PrivateAttr(default='System.Object')


ObjectModel = Object


class Deserialisers(DtoEnum):
    Tfl__Api__Presentation__Entities__StopPointCategory = StopPointCategory
    Tfl__Api__Presentation__Entities__Mode = Mode
    Tfl__Api__Presentation__Entities__PassengerFlow = PassengerFlow
    Tfl__Api__Presentation__Entities__TrainLoading = TrainLoading
    Tfl__Api__Presentation__Entities__Crowding = Crowding
    Tfl__Api__Presentation__Entities__Identifier = Identifier
    Tfl__Api__Presentation__Entities__LineGroup = LineGroup
    Tfl__Api__Presentation__Entities__LineModeGroup = LineModeGroup
    Tfl__Api__Presentation__Entities__AdditionalProperties = AdditionalProperties
    Tfl__Api__Presentation__Entities__Place = Place
    Tfl__Api__Presentation__Entities__StopPoint = StopPoint
    Tfl__Api__Presentation__Entities__LineServiceTypeInfo = LineServiceTypeInfo
    Tfl__Api__Presentation__Entities__LineSpecificServiceType = LineSpecificServiceType
    Tfl__Api__Presentation__Entities__LineServiceType = LineServiceType
    Tfl__Api__Presentation__Entities__PredictionTiming = PredictionTiming
    Tfl__Api__Presentation__Entities__Prediction = Prediction
    Tfl__Api__Presentation__Entities__ArrivalDeparture = ArrivalDeparture
    Tfl__Api__Presentation__Entities__StopPointRouteSection = StopPointRouteSection
    Tfl__Api__Presentation__Entities__DisruptedPoint = DisruptedPoint
    Tfl__Api__Presentation__Entities__StopPointsResponse = StopPointsResponse
    Tfl__Api__Presentation__Entities__SearchMatch = SearchMatch
    Tfl__Api__Presentation__Entities__SearchResponse = SearchResponse
    System__Object = Object


"""Accessed via dynamic method resolution under `tubeulator.fetch.road`.

!!! example "Example: `tubeulator.fetch.road.meta_modes()`"

    ```py
    >>> road_by_ids = fetch.road.road_by_ids(ids="A1")
    >>> print(road_by_ids[0].model_dump_json(indent=2))
    {
      "Id": "a1",
      "DisplayName": "A1",
      "Group": null,
      "StatusSeverity": "Good",
      "StatusSeverityDescription": "No Exceptional Delays",
      "Bounds": "[[-0.25616,51.5319],[-0.10234,51.6562]]",
      "Envelope": "[[-0.25616,51.5319],[-0.25616,51.6562],[-0.10234,51.6562],[-0.10234,51.5319],[-0.25616,51.5319]]",
      "StatusAggregationStartDate": null,
      "StatusAggregationEndDate": null,
      "Url": "/Road/a1"
    }
    ```
"""

from .types import RouteEnum


class RoadEndpointRoutes(RouteEnum):
    all_roads = "/"
    """Gets all roads managed by TfL"""
    meta_categories = "/Meta/Categories"
    """Gets a list of valid RoadDisruption categories"""
    meta_severities = "/Meta/Severities"
    """Gets a list of valid RoadDisruption severity codes"""
    disruption_by_ids = "/all/Disruption/{disruptionIds}"
    """Gets a list of active disruptions filtered by disruption Ids."""
    street_disruption = "/all/Street/Disruption"
    """Gets a list of disrupted streets. If no date filters are provided, current
    disruptions are returned."""
    road_by_ids = "/{ids}"
    """Gets the road with the specified id (e.g. A1)"""
    road_disruption = "/{ids}/Disruption"
    """Get active disruptions, filtered by road ids"""
    road_status = "/{ids}/Status"
    """Gets the specified roads with the status aggregated over the date range
    specified, or now until the end of today if no dates are passed."""



"""Tests for the accident_stats API."""

from accident_stats_data import all_endpoints
from pytest import mark

from tubeulator import fetch

tested_eps = {"year_accidents"}
untested_eps = set()


def test_accident_stats_endpoints():
    assert list(vars(fetch.accident_stats)) == all_endpoints
    assert set(vars(fetch.accident_stats)) == tested_eps.union(untested_eps)


@mark.skip(reason="There are 50,626 results for 2019, then none for subsequent years")
def test_year_accidents():
    year_accidents = fetch.accident_stats.year_accidents(year=2019)
    assert year_accidents



"""Tests for the search API."""

from pytest import mark
from search_data import all_endpoints

from tubeulator import fetch

tested_eps = {
    "search",
    "meta_categories",
    "meta_search_providers",
    "meta_sorts",
}
untested_eps = {
    "bus_schedules",
}


def test_search_endpoints():
    assert list(vars(fetch.search)) == all_endpoints
    assert set(vars(fetch.search)) == tested_eps.union(untested_eps)


@mark.skip(reason="Not sure how to call this one?")
def test_search():
    search = fetch.search.search()
    assert search


@mark.skip(reason="check_vars_supplied invalidates the optional argument 'query'.")
def test_bus_schedules():
    bus_schedules = fetch.search.bus_schedules(query="Canning Town")
    assert bus_schedules


def test_meta_categories():
    meta_categories = fetch.search.meta_categories()
    assert meta_categories


def test_meta_search_providers():
    meta_search_providers = fetch.search.meta_search_providers()
    assert meta_search_providers


def test_meta_sorts():
    meta_sorts = fetch.search.meta_sorts()
    assert meta_sorts



"""Tests for the place API."""

from place_data import all_endpoints
from pytest import mark

from tubeulator import fetch

tested_eps = {
    "meta_categories",
    "places_by_type",
    "place_by_id",
}
untested_eps = {
    "forward_requests",
    "places_by_geo_region",
    "meta_place_types",
    "place_by_type_at_coordinates",
    "search_places",
}


def test_place_endpoints():
    assert list(vars(fetch.place)) == all_endpoints
    assert set(vars(fetch.place)) == tested_eps.union(untested_eps)


@mark.skip(reason="404 (endpoint not found)?")
def test_places_by_geo_region():
    places_by_geo_region = fetch.place.places_by_geo_region()
    assert places_by_geo_region


def test_meta_categories():
    meta_categories = fetch.place.meta_categories()
    assert meta_categories


@mark.skip(reason="Receives list of strings but expects objects")
def test_meta_place_types():
    meta_place_types = fetch.place.meta_place_types()
    assert meta_place_types


@mark.skip(reason="404 (endpoint not found)?")
def test_search_places():
    search_places = fetch.place.search_places()
    assert search_places


@mark.parametrize("place_type", ["CarPark", "BikePoint", "ChargeConnector"])
def test_places_by_type(place_type):
    places_by_type = fetch.place.places_by_type(types=place_type)
    assert places_by_type


def test_place_by_id():
    place_by_id = fetch.place.place_by_id(id="CarParks_800503")
    assert place_by_id


@mark.skip(reason="Value is returned empty, despite coming from places_by_type result.")
def test_place_by_type_at_coordinates():
    """Car parks near Morden Station."""
    place = fetch.place.place_by_type_at_coordinates(
        type="CarPark",
        lat="51.404413",
        lon="-0.194517",
    )
    assert place



"""Tests for the occupancy API."""

from occupancy_data import all_endpoints
from pytest import mark

from tubeulator import fetch

tested_eps = {
    "bike_points",
    "charge_connector",
    "charge_connector_ids",
}
untested_eps = {
    "car_park",
    "car_park_id",
}


def test_occupancy_endpoints():
    assert list(vars(fetch.occupancy)) == all_endpoints
    assert set(vars(fetch.occupancy)) == tested_eps.union(untested_eps)


def test_bike_points():
    bike_points = fetch.occupancy.bike_points(ids="BikePoints_100")
    assert bike_points


@mark.skip(reason="500: Internal server error")
def test_car_park():
    car_park = fetch.occupancy.car_park()
    assert car_park


@mark.skip(reason="500: Internal server error")
def test_car_park_id():
    car_park_id = fetch.occupancy.car_park_id(id="CarParks_800503")
    assert car_park_id


def test_charge_connector():
    charge_connector = fetch.occupancy.charge_connector()
    assert charge_connector


def test_charge_connector_ids():
    charge_connector_ids = fetch.occupancy.charge_connector_ids(
        ids="ChargePointCM-BPT62200886-143260",
    )
    assert charge_connector_ids



from typing import TypeVar

import patito as pt

from ..utils.paths import stationdata_detailed_path
from .data_model import (
    ModeAndLine,
    Platform,
    PlatformService,
    Station,
    StationPoint,
)

__all__ = [
    "load_model",
    "load_lines",
    "load_platforms",
    "load_platform_services",
    "load_stations",
    "load_station_points",
]

M = TypeVar("M", bound=pt.Model)


def load_model(model_class: M) -> pt.DataFrame:
    """Load the CSV file associated with a patito model `model_class`."""
    source_csv = stationdata_detailed_path / model_class._source.default
    dataset = model_class.DataFrame.read_csv(source_csv)
    dataset = dataset[model_class.columns]
    dataset.validate()
    return dataset


def load_lines() -> pt.DataFrame:
    return load_model(ModeAndLine)


def load_platforms() -> pt.DataFrame:
    return load_model(Platform)


def load_platform_services() -> pt.DataFrame:
    return load_model(PlatformService)


def load_stations() -> pt.DataFrame:
    return load_model(Station)


def load_station_points() -> pt.DataFrame:
    return load_model(StationPoint)



import polars as pl

from tubeulator.topology.combine import load_lines_by_station, load_stations_by_line

__all__ = ["test_load_stations_by_line", "test_load_lines_by_station"]


def test_load_stations_by_line():
    line2stns = load_stations_by_line()
    stn_counts = line2stns.with_columns(
        pl.col("StationName").list.len().alias("Count"),
    ).sort("Count")
    assert stn_counts["Count"].min() == 2
    assert stn_counts["Count"].max() == 112


def test_load_lines_by_station():
    stn2lines = load_lines_by_station()
    line_counts = stn2lines.with_columns(
        pl.col("Lines").list.len().alias("Count"),
    ).sort("Count")
    assert line_counts["Count"].min() == 1
    assert line_counts["Count"].max() == 7



"""Tests for the air_quality API."""

from air_quality_data import all_endpoints
from pytest import mark

from tubeulator import fetch

tested_eps = {"air_quality_data"}
untested_eps = set()


def test_air_quality_endpoints():
    assert list(vars(fetch.air_quality)) == all_endpoints
    assert set(vars(fetch.air_quality)) == tested_eps.union(untested_eps)


@mark.skip(reason="Returns an empty Object")
def test_air_quality_data():
    air_quality_data = fetch.air_quality.air_quality_data()
    assert air_quality_data.model_dump()



import json
from functools import cache
from pathlib import Path

from pydantic import BaseModel

from .paths import openapi_schemas_path, openapi_unified_path

__all__ = [
    "find_schema",
    "find_schema_by_name",
    "load_endpoint_schema",
    "load_endpoint_component_schemas",
    "unified_api_schema",
]


def find_schema(schema_dir: Path, match: str = "*", levels: int = 0) -> Path:
    """Get the next schema after traversing the given number of levels (default: no levels,
    i.e. directly in the given directory). The schema must have a ``.json`` suffix.
    """
    try:
        glob_prefix = "*/" * levels
        glob_pattern = f"{glob_prefix}{match}.json"
        return next(schema_dir.glob(glob_pattern))
    except StopIteration:
        raise ValueError(f"No JSON file {levels=} below {schema_dir}")


@cache
def find_schema_by_name(schema_name: str) -> Path:
    return find_schema(schema_dir=openapi_schemas_path, match=schema_name, levels=1)


@cache
def load_endpoint_schema(schema_name: str) -> dict:
    """Load an entire JSON schema for an API endpoint by its name, e.g. "Line" or "Mode"."""
    schema_file = find_schema_by_name(schema_name)
    endpoint_schema = json.loads(schema_file.read_text())
    return endpoint_schema


class SchemaComponents(BaseModel):
    schemas: dict = {}


class Schema(BaseModel):
    components: SchemaComponents


@cache
def load_endpoint_component_schemas(schema_name: str) -> dict[str, dict]:
    """Load all component schemas of a JSON schema for an API endpoint by endpoint name."""
    schema_file = find_schema_by_name(schema_name)
    schema_json = schema_file.read_text()
    endpoint_schema = Schema.model_validate_json(schema_json)
    component_schemas = endpoint_schema.components.schemas
    return component_schemas


unified_api_schema = find_schema(schema_dir=openapi_unified_path, levels=1)



from __future__ import annotations

from enum import Enum

import patito as pt
from patito import Model
from pydantic import PrivateAttr

__all__ = [
    "StationPoint",
    "Station",
    "Platform",
    "PlatformService",
    "ModeAndLine",
    "CardinalDirectionEnum",
]


class StationPoint(Model):
    _source: str = PrivateAttr("StationPoints.csv")
    UniqueId: str
    StationUniqueId: str  # corresponds to UniqueId in the Station model
    AreaName: str
    AreaId: int
    Level: int
    Lat: float
    Lon: float
    FriendlyName: str


class Station(Model):
    _source: str = PrivateAttr("Stations.csv")
    StationUniqueId: str = pt.Field(derived_from="UniqueId")
    StationName: str = pt.Field(derived_from="Name")
    FareZones: str  # |-separated list[str]
    HubNaptanCode: str | None
    Wifi: bool
    OutsideStationUniqueId: str
    # Other fields not included here...


class Platform(Model):
    _source: str = PrivateAttr("Platforms.csv")
    PlatformUniqueId: str = pt.Field(derived_from="UniqueId")
    StationUniqueId: str
    PlatformNumber: str | None
    CardinalDirection: str | None  # CardinalDirectionEnum
    PlatformNaptanCode: str | None
    PlatformFriendlyName: str = pt.Field(derived_from="FriendlyName")
    IsCustomerFacing: bool
    HasServiceInterchange: bool
    # Other fields not included here...


class PlatformService(Model):
    _source: str = PrivateAttr("PlatformServices.csv")
    PlatformUniqueId: str
    StopAreaNaptanCode: str
    Line: str  # corresponds to Name in the ModesAndLines model
    DirectionTowards: str | None
    # Other fields not included here...
    PlatformServiceGroupName: str | None = pt.Field(derived_from="GroupName")


class ModeAndLine(Model):
    _source: str = PrivateAttr("ModesAndLines.csv")
    Mode: str
    Name: str


class CardinalDirectionEnum(Enum):
    Westbound = "WB"
    Eastbound = "EB"
    Northbound = "NB"
    Southbound = "SB"



import tubeulator


def test_identifier_parse():
    """This was a problem case due to the key and the model class name matching."""
    dto = tubeulator.generated.Journey.Identifier
    data = {"crowding": {"passengerFlows": [], "trainLoadings": []}}
    dto.model_validate(data)



"""Deprecated methods, GTFS data isn't very useful on its own."""

from pathlib import Path

import patito as pt

from ..utils.paths import stationdata_gtfs_path
from .data_model import GTFSStop

__all__ = ["load_transport_topology", "load_stations"]


def load_transport_topology(
    source: Path = stationdata_gtfs_path / "stops.txt",
) -> pt.DataFrame:
    stops = GTFSStop.DataFrame.read_csv(source)
    stops.validate()
    return stops


def load_stations() -> pt.DataFrame:
    """Load a DataFrame of ID and name for all stations.

    The stop_id and stop_code are identical for the stations."""
    topology = load_transport_topology()
    stations = topology.filter(pt.col("location_type") == 1)
    station_columns = ["stop_id", "stop_name"]  # stations.columns[:3]
    return stations[station_columns]


def load_stops() -> pt.DataFrame:
    """Load a DataFrame of ID and name for all stops.

    The stop_id and stop_code are not identical for the stops."""
    topology = load_transport_topology()
    stations = topology.filter(pt.col("location_type") != 1)
    return stations
    # station_columns = ["stop_id", "stop_name"]  # stations.columns[:3]
    # return stations[station_columns]



"""Accessed via dynamic method resolution under `tubeulator.fetch.accident_stats`.

!!! example "Example: `tubeulator.fetch.accident_stats.year_accidents()`"

    ```py
    >>> accidents = fetch.accident_stats.year_accidents(year=2019)
    >>> print(accidents[0].model_dump_json(indent=2))
    {
      "Id": 345979,
      "Lat": 51.570865,
      "Lon": -0.231959,
      "Location": "On Edgware Road Near The Junction With north Circular Road",
      "Date": "2019-01-04T21:22:00Z",
      "Severity": "Slight",
      "Borough": "Barnet",
      "Casualties": [
        {
          "Age": 20,
          "Class": "Driver",
          "Severity": "Slight",
          "Mode": "PoweredTwoWheeler",
          "AgeBand": "Adult"
        }
      ],
      "Vehicles": [
        {
          "Type": "Motorcycle_500cc_Plus"
        },
        {
          "Type": "Car"
        }
      ]
    }
    ```
"""

from .types import RouteEnum


class AccidentStatsEndpointRoutes(RouteEnum):
    year_accidents = "/{year}"
    """Gets all accident details for accidents occuring in the specified year"""



"""Tests for the bike_point API."""

from bike_point_data import all_endpoints
from pytest import mark

from tubeulator import fetch

tested_eps = {"all_locations", "bike_point_by_id"}
untested_eps = {"search_stations"}


def test_bike_point_endpoints():
    assert list(vars(fetch.bike_point)) == all_endpoints
    assert set(vars(fetch.bike_point)) == tested_eps.union(untested_eps)


def test_all_locations():
    all_locations = fetch.bike_point.all_locations()
    assert all_locations


@mark.skip(reason="404 (endpoint not found)?")
def test_search_stations():
    search_stations = fetch.bike_point.search_stations()
    assert search_stations


def test_bike_point_by_id():
    bike_point_by_id = fetch.bike_point.bike_point_by_id(id="BikePoints_100")
    assert bike_point_by_id



from tubeulator.topology.load import (
    load_lines,
    load_platform_services,
    load_platforms,
    load_station_points,
    load_stations,
)

__all__ = [
    "test_load_lines",
    "test_load_platforms",
    "test_load_platform_services",
    "load_stations",
    "load_station_points",
]


def test_load_lines():
    load_lines()


def test_load_platforms():
    load_platforms()


def test_load_platform_services():
    load_platform_services()


def test_load_stations():
    load_stations()


def test_load_station_points():
    load_station_points()



from inline_snapshot import snapshot

all_endpoints = snapshot(
    [
        "within_radius",
        "forward_requests",
        "meta_categories",
        "meta_modes",
        "meta_stop_types",
        "mode",
        "mode_disruption",
        "search",
        "search_query",
        "service_types",
        "sms_id",
        "type",
        "type_page",
        "stop_point_ids",
        "stop_point_disruption",
        "arrival_departures",
        "arrivals",
        "can_reach_on_line",
        "crowding",
        "direction_to",
        "route",
        "place_types",
        "car_parks",
        "taxi_ranks",
    ],
)
